# pylint: disable=unused-import
from .post_process_rules.MustCheckReturnCodeRule import MustCheckReturnCodeRule
from .post_process_rules.UnusedFunctionsRule import UnusedFunctionsRule
from .post_process_rules.UniformAssertsConstants import UniformAssertsConstants
from .post_process_rules.MustCheckOverflow import MustCheckOverflow
from .post_process_rules.MustCheckCallerAddress import MustCheckCallerAddress
from .post_process_rules.StorageVarCollisionRule import StorageVarCollisionRule
from .post_process_rules.ImplicitImportRule import ImplicitImportRule
from .post_process_rules.UnenforcedViewRule import UnenforcedViewRule



from typing import Dict, List

from amarna.rules.gatherer_rules.RValueFunctionCallsGatherer import (
    RValueFunctionCallsGatherer,
)

from amarna.Result import Result, create_result_token

# pylint: disable=too-few-public-methods
class MustCheckCallerAddress:
    """
    Gather function calls to get_caller_address.
    """

    RULE_TEXT = (
        "The function get_caller_address returns 0"
        + " when the contract is called directly which might be unexpected"
    )
    IGNORED_RULE_TEXT = "The result of get_caller_address is ignored."

    RULE_NAME = "must-check-caller-address"

    # dictionary with function_name : index of the caller address
    TARGET_FUNCTION = {"get_caller_address": 0}

    def run_rule(self, gathered_data: Dict) -> List[Result]:

        function_calls = gathered_data[RValueFunctionCallsGatherer.GATHERER_NAME]

        results: List[Result] = []
        for call in function_calls:
            file_name, function_name, returned_list = call

            if function_name in self.TARGET_FUNCTION:
                idx = self.TARGET_FUNCTION[function_name]

                must_check = returned_list[idx].children[0]

                if must_check == "_":
                    # the caller address is being ignored
                    rule = self.IGNORED_RULE_TEXT
                else:
                    # here we say that it must be checked
                    rule = self.RULE_TEXT

                sarif = create_result_token(
                    file_name,
                    self.RULE_NAME,
                    rule,
                    must_check,
                )
                results.append(sarif)

        return results



from dataclasses import dataclass
from typing import List

from lark import Tree
from amarna.rules.GenericGatherer import GenericGatherer
from amarna.Result import token_positions, PositionType


@dataclass
class ImportType:
    """Represents an import statement."""

    import_name: str
    alias_name: str
    where_imported: str
    location: PositionType
    file_location: str


class ImportGatherer(GenericGatherer):
    """
    Gather all import statements.
    """

    GATHERER_NAME = "ImportGatherer"

    def __init__(self) -> None:
        super().__init__()
        self.import_stmts: List[ImportType] = []

    def get_gathered_data(self) -> List[ImportType]:
        return self.import_stmts

    def cairo_file(self, tree: Tree) -> None:
        # iterate over the imports
        for code_imports in tree.find_data("code_element_import"):
            where_imported = ""
            for child in code_imports.children:
                # gather from where it is being imported
                if child.data == "identifier":
                    where_imported = "/".join(map(str, child.children)) + ".cairo"

                # gather what is being imported
                if child.data == "aliased_identifier":
                    assert where_imported
                    # this is just a simple import
                    if len(child.children) == 1:
                        import_name = child.children[0].children[0]
                        stmt = ImportType(
                            str(import_name),
                            "",
                            where_imported,
                            token_positions(import_name),
                            self.fname,
                        )
                        self.import_stmts.append(stmt)

                    # this is an aliased import import XX as YY, we keep the new name
                    elif len(child.children) == 2:
                        import_name = child.children[0].children[0]
                        alias_name = child.children[1].children[0]

                        stmt = ImportType(
                            str(import_name),
                            alias_name,
                            where_imported,
                            token_positions(import_name),
                            self.fname,
                        )
                        self.import_stmts.append(stmt)



from collections import defaultdict
from typing import Dict, List
from amarna.Result import ResultMultiplePositions, result_multiple_positions

from amarna.rules.gatherer_rules.DeclaredFunctionsGatherer import (
    DeclaredFunctionsGatherer,
    FunctionType,
)
from amarna.rules.gatherer_rules.ImportGatherer import ImportGatherer, ImportType


class StorageVarCollisionRule:
    """
    Storage variable is declared multiple times across files.
    """

    RULE_TEXT = "[This](0) storage variable is also declared [here](1), shadowing it and causing both storage variables to always be equal."
    RULE_NAME = "storage-var-collision"

    def run_rule(self, gathered_data: Dict) -> List[ResultMultiplePositions]:
        declared_functions: List[FunctionType] = gathered_data[
            DeclaredFunctionsGatherer.GATHERER_NAME
        ]

        import_stmts: List[ImportType] = gathered_data[ImportGatherer.GATHERER_NAME]

        results = []
        seen = defaultdict(list)
        for func in declared_functions:
            if "storage_var" in func.decorators:
                seen[(func.file_location, func.name)].append(func)

                files_marked = []
                for imp in import_stmts:
                    # check if there is any import from the location
                    # of the storage variable
                    if func.file_location.endswith(imp.where_imported):
                        # do not flag the same file repeatedly
                        if imp.file_location in files_marked:
                            continue

                        files_marked.append(imp.file_location)
                        seen[(imp.file_location, func.name)].append(func)

        for name in seen:
            if len(seen[name]) > 1:
                sorted_seen = sorted(seen[name], key=lambda x: x.file_location)

                result = result_multiple_positions(
                    [s.file_location for s in sorted_seen],
                    self.RULE_NAME,
                    self.RULE_TEXT,
                    [s.position for s in sorted_seen],
                )
                # do not add findings twice
                if result.position_list not in (r.position_list for r in results):
                    results.append(result)

        return results



from typing import Dict, Optional

from lark import Tree

from amarna.rules.GenericGatherer import GenericGatherer

FunctionReturningType = Dict[str, int]


class FunctionsReturningErrorsGatherer(GenericGatherer):
    GATHERER_NAME = "FunctionsReturningErrorsGatherer"

    def __init__(self) -> None:
        super().__init__()
        self.functions_returning_errors: FunctionReturningType = {}

    def get_gathered_data(self) -> FunctionReturningType:
        return self.functions_returning_errors

    def code_element_function(self, tree: Tree) -> None:
        return_code_pos = -1
        function_name: Optional[str] = None
        for child in tree.children:
            if child.data == "identifier_def":
                function_name = str(child.children[0])

            elif child.data == "identifier_list":
                for idx, args in enumerate(child.find_data("identifier_def")):
                    return_name = args.children[0]

                    if "return_code" in return_name or "success" in return_name:
                        return_code_pos = idx

        if return_code_pos == -1:
            return

        assert function_name

        self.functions_returning_errors[function_name] = return_code_pos



from typing import List
from lark import Tree
from amarna.Result import PositionType, getPosition
from amarna.rules.GenericGatherer import GenericGatherer
from dataclasses import dataclass


@dataclass
class EventEmitType:
    """Represents an event emit (a function call of the emit method for a @event function)."""

    event_name: str
    event_emit_position: PositionType


@dataclass
class FunctionEmittingEvent:
    """A function that emits events, and their locations."""

    function_name: str
    function_position: PositionType
    function_location: str
    events_emitted_list: List[EventEmitType]


class FunctionsEmittingEventsGatherer(GenericGatherer):
    GATHERER_NAME = "FunctionsEmittingEvents"

    def __init__(self) -> None:
        super().__init__()
        self.functions_emitting_events: List[FunctionEmittingEvent] = []

    def get_gathered_data(self) -> List[FunctionEmittingEvent]:
        return self.functions_emitting_events

    def code_element_function(self, tree: Tree) -> None:
        for child in tree.children:
            if child.data == "identifier_def":
                function_name = str(child.children[0])
                break

        events_emitted_list: List[EventEmitType] = []

        for call in tree.find_data("function_call"):
            if len(call.children) > 0:
                ids = call.children[0]
                assert ids.data == "identifier"
                if len(ids.children) == 2 and ids.children[1] == "emit":
                    event_name = ids.children[0]
                    event_position = getPosition(ids)
                    events_emitted_list.append(EventEmitType(event_name, event_position))

        if events_emitted_list:
            event = FunctionEmittingEvent(
                function_name, getPosition(tree), self.fname, events_emitted_list
            )
            self.functions_emitting_events.append(event)



from lark import Tree

from amarna.rules.GenericRule import GenericRule
from amarna.Result import create_result, getPosition


class ArithmeticOperationsRule(GenericRule):
    """
    Check arithmetic operations:
        - reports ALL multiplications and divisions
        - reports ONLY addition and subtraction that do not involve a register like [ap - 1]
    """

    RULE_TEXT = "Cairo arithmetic is defined over a finite field and has potential for overflows."
    RULE_PREFIX = "arithmetic-"


class MulArithmeticOperationsRule(ArithmeticOperationsRule):
    RULE_NAME = ArithmeticOperationsRule.RULE_PREFIX + "mul"

    def expr_mul(self, tree: Tree) -> None:
        result = create_result(self.fname, self.RULE_NAME, self.RULE_TEXT, getPosition(tree))
        self.results.append(result)


class DivArithmeticOperationsRule(ArithmeticOperationsRule):
    RULE_NAME = ArithmeticOperationsRule.RULE_PREFIX + "div"

    def expr_div(self, tree: Tree) -> None:
        result = create_result(self.fname, self.RULE_NAME, self.RULE_TEXT, getPosition(tree))
        self.results.append(result)


class AddArithmeticOperationsRule(ArithmeticOperationsRule):
    RULE_NAME = ArithmeticOperationsRule.RULE_PREFIX + "add"

    def expr_add(self, tree: Tree) -> None:
        # ignore adding to registers
        if tree.children[0].data == "atom_reg":
            return

        result = create_result(self.fname, self.RULE_NAME, self.RULE_TEXT, getPosition(tree))
        self.results.append(result)


class SubArithmeticOperationsRule(ArithmeticOperationsRule):
    RULE_NAME = ArithmeticOperationsRule.RULE_PREFIX + "sub"

    def expr_sub(self, tree: Tree) -> None:
        # ignore subtracting to registers
        if tree.children[0].data == "atom_reg":
            return

        result = create_result(self.fname, self.RULE_NAME, self.RULE_TEXT, getPosition(tree))
        self.results.append(result)



import os
import argparse
from amarna.amarna import Amarna, analyze_directory, analyze_file
from amarna.Result import Result, ResultMultiplePositions, output_result
from amarna.Result import SARIF_MODE, SUMMARY_MODE
from typing import List, Union

example_usage = """---------------\nUsage examples\n---------------
Analyze a Cairo project in the current directory and export results to a file:
 amarna . -o out.sarif

Analyze a single file `deleverage.cairo` and export results to a file:
 amarna deleverage.cairo -o deleverage.sarif

Analyze a single file `code.cairo` and print a summary of the results:
 amarna code.cairo -s

Parse a Cairo file and output the recovered AST in `png`:
 amarna file.cairo -png

Parse a Cairo file and only run the unused_import rule:
 amarna file.cairo --rules=unused-imports

Analyze a Cairo file using all rules except the arithmetic-add rule:
 amarna file.cairo --except-rules=arithmetic-add
 """


def parse_comma_sep_strings(s: str) -> List[str]:
    if s:
        return [token.strip() for token in s.split(",") if token.strip()]
    else:
        return []


def get_rule_names(rules: str, excluded: str) -> List[str]:
    ALL_RULES = Amarna.get_all_rule_names()

    rules = parse_comma_sep_strings(rules)
    excluded = parse_comma_sep_strings(excluded)

    for rule in rules + excluded:
        if rule not in ALL_RULES:
            print("Unknown rule: " + repr(rule))
            exit(-1)

    if rules:
        base_rules = rules
    else:
        base_rules = ALL_RULES

    return [rule for rule in base_rules if rule not in excluded]


def filter_results_from_disable(
    results: List[Union[Result, ResultMultiplePositions]]
) -> List[Union[Result, ResultMultiplePositions]]:
    first_lines_per_file = {}
    disable_token = "# amarna: disable="

    new_results = []
    for result in results:
        if isinstance(result, Result):
            filename = result.filename
        elif isinstance(result, ResultMultiplePositions):
            filename = result.filenames[0]
        if filename in first_lines_per_file:
            first_line = first_lines_per_file[filename]
        else:
            with open(filename, "r", encoding="utf8") as f:
                first_line = f.readline().strip()
                first_lines_per_file[filename] = first_line

        if not first_line.startswith(disable_token):
            new_results.append(result)
            continue

        rule_tok = first_line.split(disable_token)[1]
        rule_list = get_rule_names(None, rule_tok)

        if result.rule_name in rule_list:
            new_results.append(result)

    return new_results


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Amarna is a static-analyzer for the Cairo programming language.",
        epilog=example_usage,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "filename",
        metavar="-f",
        type=str,
        help="the name of the .cairo file or directory with .cairo files to analyze",
    )

    parser.add_argument("-p", "--print", action="store_true", help="print output")

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="file to write the output results in sarif format",
    )

    parser.add_argument("-s", "-summary", "--summary", action="store_true", help="output summary")

    parser.add_argument(
        "-png", "--png", action="store_true", help="save a png with the AST of a file"
    )

    parser.add_argument(
        "-rules",
        "--rules",
        type=str,
        help="Only run this set of rules. Enter rule names comma-separated, e.g., dead-store,unused-arguments",
    )

    parser.add_argument(
        "-exclude-rules",
        "--exclude-rules",
        type=str,
        help="Exclude these rules from the analysis. Enter rule names comma-separated, e.g., dead-store,unused-arguments",
    )

    parser.add_argument(
        "-show-rules",
        "--show-rules",
        action="store_true",
        help="Show all supported rules and descriptions.",
    )

    parser.add_argument(
        "-disable-inline",
        "--disable-inline",
        action="store_true",
        help="Disable rules with inline comments. The comments should be the first line and of the form: # amarna: disable=rulename1,rulename2",
    )

    args = parser.parse_args()

    if args.show_rules:
        Amarna.print_rule_names_and_descriptions()
        return 1

    filename = args.filename

    if not os.path.isabs(filename):
        filename = os.path.join(os.getcwd(), filename)

    rule_set_names: List[str] = get_rule_names(args.rules, args.exclude_rules)

    results: List[Union[Result, ResultMultiplePositions]]
    if os.path.isdir(filename):
        results = analyze_directory(filename, rule_set_names)
    else:
        results = analyze_file(filename, rule_set_names, png=args.png)

    if args.disable_inline:
        results = filter_results_from_disable(results)

    mode = SARIF_MODE
    if args.summary:
        args.print = True
        mode = SUMMARY_MODE

    if args.output or args.print:
        output_result(results, args.output, args.print, mode)


if __name__ == "__main__":
    main()



from typing import List

from lark import Tree, Visitor


class GenericRule(Visitor):
    """
    Generic class for a runnable rule.
    """

    def run_rule(self, fname: str, tree: Tree) -> List:
        # TODO (montyly): investigate if this pylint rule should be removed
        # and an __init__ function created
        # pylint: disable=attribute-defined-outside-init
        self.fname = fname
        self.results: List = []
        self.original_tree = tree
        self.visit(tree)
        return self.results



import inspect
import os
from typing import Any, List, Dict, Union
from pathlib import Path
from lark import Lark, Tree, tree, exceptions
from amarna.Result import Result, ResultMultiplePositions

from amarna.rules import gatherer_rules_module, local_rules_module, post_process_rules_module
from amarna.rules.GenericRule import GenericRule


def make_png(t: tree.Tree, out_name: str) -> None:
    """
    Creates a PNG of the AST in the out_name file
    """
    tree.pydot__tree_to_png(t, out_name)


class Amarna:
    """
    The main class -- loads the cairo grammar, and runs rules.
    """

    @staticmethod
    def load_cairo_grammar() -> Lark:
        _module_dir = Path(__file__).resolve().parent
        GRAMMAR_FILENAME = _module_dir.joinpath("grammars", "cairo.lark")
        with open(GRAMMAR_FILENAME, "r", encoding="utf8") as f:
            buf = f.read()
        cairo_parser = Lark(
            buf,
            start=[
                "cairo_file",
                "code_block",
                "code_element",
                "expr",
                "instruction",
                "type",
                "typed_identifier",
            ],
            parser="lalr",
            propagate_positions=True,
        )

        return cairo_parser

    @staticmethod
    def load_classes_in_module(module: Any) -> List:
        return [cls() for (_, cls) in inspect.getmembers(module, inspect.isclass)]

    @staticmethod
    def get_all_rule_names() -> List[str]:
        return list(
            map(
                lambda x: x.RULE_NAME,
                Amarna.load_classes_in_module(local_rules_module)
                + Amarna.load_classes_in_module(post_process_rules_module),
            )
        )

    @staticmethod
    def print_rule_names_and_descriptions() -> List[str]:
        ruleset = Amarna.load_classes_in_module(local_rules_module) + Amarna.load_classes_in_module(
            post_process_rules_module
        )
        for rule in ruleset:
            print(rule.RULE_NAME)

    def __init__(self, only_these_analysis_rules: List[str]) -> None:
        """
        Load Cairo grammar and analysis rules.
        """
        self.parser = Amarna.load_cairo_grammar()
        # self.data maps the gather name to the value returned by
        # the different "gather" functions
        # Which return a GenericGatherType type (python generic)
        self.data: Dict[str, Any] = {}
        self.rules: List[GenericRule] = [
            rule
            for rule in Amarna.load_classes_in_module(local_rules_module)
            if rule.RULE_NAME in only_these_analysis_rules
        ]
        self.post_process_rules = [
            rule
            for rule in Amarna.load_classes_in_module(post_process_rules_module)
            if rule.RULE_NAME in only_these_analysis_rules
        ]
        self.gatherers = Amarna.load_classes_in_module(gatherer_rules_module)

    def run_local_rules(
        self, filename: str, parsed_cairo_file: Any, png: bool = False
    ) -> List[Union[Result, ResultMultiplePositions]]:
        """
        Run all local rules.

        """
        if png:
            png_filename = os.path.basename(filename).split(".")[0] + ".png"
            make_png(parsed_cairo_file, png_filename)

        results = []
        for Rule in self.rules:
            results += Rule.run_rule(filename, parsed_cairo_file)

        return results

    def run_gatherer_rules(self, filename: str, parsed_cairo_file: Tree) -> None:
        """
        Run all gatherer rules.
        """
        for Gatherer in self.gatherers:
            self.data[Gatherer.GATHERER_NAME] = Gatherer.gather(filename, parsed_cairo_file)

    def run_post_process_rules(self) -> List[Union[Result, ResultMultiplePositions]]:
        """
        Run all post-processing rules.
        """
        results = []
        for Rule in self.post_process_rules:
            results += Rule.run_rule(self.data)
        return results

    def parse_cairo_file(self, filename: str) -> Union[Tree, None]:
        """
        Parse the cairo file
        """
        try:
            with open(filename, "r", encoding="utf8") as f:
                # the cairo grammar requires a newline at the end
                return self.parser.parse(f.read() + "\n", start="cairo_file")
        except (exceptions.UnexpectedCharacters, exceptions.UnexpectedToken) as e:
            print(f"Could not parse {filename}: {e}")
            return None


def analyze_directory(rootdir: str, rule_names: List[str]) -> List[Any]:
    """
    Run rules in all .cairo files inside a directory.
    """
    amarna = Amarna(rule_names)

    all_results = []
    run_gather = False

    for subdir, _dirs, files in os.walk(rootdir):
        for file in files:
            fname = os.path.join(subdir, file)

            if fname.endswith(".cairo"):
                parsed_cairo_file = amarna.parse_cairo_file(fname)
                if not parsed_cairo_file:
                    continue

                all_results += amarna.run_local_rules(fname, parsed_cairo_file)
                amarna.run_gatherer_rules(fname, parsed_cairo_file)
                run_gather = True

    if run_gather:
        all_results += amarna.run_post_process_rules()
    else:
        print(f"Could not find any cairo files in {rootdir}")
    return sorted(all_results, key=lambda x: x.to_summary())


def analyze_file(
    fname: str, rules_names: List[str], png: bool = False
) -> List[Union[Result, ResultMultiplePositions]]:
    """
    Run analysis rules on a .cairo file.
    """
    amarna = Amarna(rules_names)
    all_results = []
    parsed_cairo_file = amarna.parse_cairo_file(fname)
    if parsed_cairo_file:
        all_results += amarna.run_local_rules(fname, parsed_cairo_file, png)
        amarna.run_gatherer_rules(fname, parsed_cairo_file)
        all_results += amarna.run_post_process_rules()
    return sorted(all_results, key=lambda x: x.to_summary())


if __name__ == "__main__":
    pass



from typing import List, Tuple, Any

from lark import Tree


from amarna.rules.GenericGatherer import GenericGatherer


class RValueFunctionCallsGatherer(GenericGatherer):
    """
    Gather function calls that return to a variable, and their return values.
    """

    GATHERER_NAME = "RValueFunctionCallsGatherer"

    def __init__(self) -> None:
        super().__init__()
        self.function_calls: List[Tuple[str, str, Any]] = []

    def get_gathered_data(self) -> List[Tuple[str, str, Any]]:
        return self.function_calls

    def code_element_reference(self, tree: Tree) -> None:

        children = tree.children
        if len(children) != 2 or children[1].data != "rvalue_expr":
            return

        rvalues = children[1]
        returned_values = children[0]

        returned_list = list(returned_values.find_data("identifier_def"))

        for func in rvalues.find_data("function_call"):
            func_id = func.children[0]
            # TODO (montyly): mypy complain that the next element has no attribute value
            function_name = func_id.children[0].value  # type: ignore
            tup = (self.fname, function_name, returned_list)
            self.function_calls.append(tup)



from typing import Set, Optional, Any
from lark import Tree, Token

from amarna.Result import PositionType, create_result
from amarna.rules.GenericRule import GenericRule


class UnusedArgumentRule(GenericRule):
    """
    Check for unused arguments inside a function
    """

    RULE_TEXT = "Unused arguments might indicate a misspelled variable use or unnecessary argument."
    RULE_NAME = "unused-arguments"

    # visit the code_element_function node of the AST
    def code_element_function(self, tree: Tree) -> None:
        # pylint: disable=too-many-branches

        function_name: Optional[str] = None
        code_block: Optional[Any] = None

        arguments = set()
        for child in tree.children:
            # get the function name
            if child.data == "identifier_def":
                # TODO (montyly): attribute access error
                function_name = child.children[0]  # type: ignore

            # get the arguments names
            elif child.data == "arguments":
                for args in child.find_data("identifier_def"):
                    arguments.add(args.children[0])

            # get the function code ASTree
            elif child.data == "code_block":
                code_block = child

            # ignore if the function is a @storage_var
            elif child.data == "decorator_list":
                for decorator in child.find_data("identifier_def"):
                    # since this is a class attribute, there is no need
                    # to check argument usage because there is no code.
                    if decorator.children[0] == "storage_var":
                        return

        assert function_name
        assert code_block

        # find used identifiers in the function code
        used_ids: Set[Token] = set()
        for code_child in code_block.find_data("identifier"):
            used_ids.add(code_child.children[0])

        unused_arguments = arguments - used_ids
        if not unused_arguments:
            return

        # if NO argument is used, guess that its due to the cast pattern
        # TODO: make precise by checking usage of cast(fp - 2 - STRUCT.SIZE, STRUCT*) and check that
        #       - struct name matches the casted type
        #       - all argument types match the struct types and order
        if len(unused_arguments) == len(arguments) and function_name.endswith("_new"):
            return

        # # gather all hint code and check if the arguments are used there
        # all_hints = ""
        # for hint in self.original_tree.find_data("code_element_hint"):
        #     all_hints += hint.children[0]

        # # remove imports used in hints
        # used_in_hints = set()
        # for unused in unused_arguments:
        #     argument_hint = "ids." + unused.value
        #     if argument_hint in all_hints:
        #         used_in_hints.add(unused)

        # unused_arguments = unused_arguments - used_in_hints

        # find if the current tree is part of a @contract_interface
        # to ignore unused arguments in that case
        for struct in self.original_tree.find_data("code_element_struct"):
            for child in struct.find_data("decorator_list"):
                for decorator in child.find_data("identifier_def"):
                    if decorator.children[0] == "contract_interface":
                        if tree in struct.iter_subtrees():
                            return

        for child in tree.children:
            if child.data == "decorator_list":
                for args in child.find_data("identifier_def"):
                    decorator = args.children[0]
                    if decorator in [
                        "event",
                    ]:
                        return

        # report unused arguments
        for arg in sorted(unused_arguments):
            # TODO (montyly): mypy complain about the next attributes accesses
            positions = PositionType(arg.line, arg.column, arg.end_line, arg.end_column)  # type: ignore
            sarif = create_result(
                self.fname,
                self.RULE_NAME,
                self.RULE_TEXT,
                positions,
            )
            self.results.append(sarif)



from typing import Dict, List
from amarna.Result import Result, create_result

from amarna.rules.gatherer_rules.DeclaredFunctionsGatherer import (
    DeclaredFunctionsGatherer,
    FunctionType,
)
from amarna.rules.gatherer_rules.AllFunctionCallsGatherer import (
    AllFunctionCallsGatherer,
    FunctionCallType,
)
from amarna.rules.gatherer_rules.FunctionsUsedAsCallbacksGatherer import (
    CallbackFunctionType,
    FunctionsUsedAsCallbacksGatherer,
)
from amarna.rules.gatherer_rules.ImportGatherer import ImportGatherer, ImportType

# pylint: disable=too-few-public-methods
class UnusedFunctionsRule:
    """
    Match declared functions and called functions or callbacks to find unused functions.
    """

    RULE_TEXT = "This function is never called."
    RULE_NAME = "unused-function"

    # TODO: handle import shadowing other function names
    UNUSED_FP_DECORATORS = ["external", "view", "constructor", "l1_handler"]

    def run_rule(self, gathered_data: Dict) -> List[Result]:
        declared_functions: List[FunctionType] = gathered_data[
            DeclaredFunctionsGatherer.GATHERER_NAME
        ]
        function_calls: List[FunctionCallType] = gathered_data[
            AllFunctionCallsGatherer.GATHERER_NAME
        ]

        callbacks: List[CallbackFunctionType] = gathered_data[
            FunctionsUsedAsCallbacksGatherer.GATHERER_NAME
        ]

        import_stmts: List[ImportType] = gathered_data[ImportGatherer.GATHERER_NAME]

        results = []

        all_called = []
        for call in function_calls:
            all_called.append(call.function_name)

        for call in callbacks:
            all_called.append(call.function_name)

        # replace with actual function name if alias
        for imp in import_stmts:
            if not imp.alias_name:
                continue
            if imp.alias_name in all_called:
                all_called.remove(imp.alias_name)
                all_called.append(imp.import_name)

        for func in declared_functions:
            # ignore if it was called
            if func.name in all_called:
                continue

            # ignore cairo standard lib functions
            if "starkware/cairo/common/" in func.file_location:
                continue

            # ignore if it has a decorator that means it is not supposed to be called directly
            if any(decorator in self.UNUSED_FP_DECORATORS for decorator in func.decorators):
                continue

            result = create_result(
                func.file_location,
                self.RULE_NAME,
                self.RULE_TEXT,
                func.position,
            )
            results.append(result)

        return results



from dataclasses import dataclass
from typing import Optional, List

from lark import Tree
from amarna.rules.GenericGatherer import GenericGatherer
from amarna.Result import getPosition, PositionType


@dataclass
class FunctionType:
    """Represents a function declaration."""

    name: str
    position: PositionType
    file_location: str
    decorators: List[str]


class DeclaredFunctionsGatherer(GenericGatherer):
    """
    Gather all declared functions.
    """

    GATHERER_NAME = "DeclaredFunctionsGatherer"

    def __init__(self) -> None:
        super().__init__()
        self.declared_functions: List[FunctionType] = []

    def get_gathered_data(self) -> List[FunctionType]:
        return self.declared_functions

    def code_element_function(self, tree: Tree) -> None:
        function_name: Optional[str] = None

        # find if the current tree is part of a @contract_interface
        # to ignore if unused in that case
        for struct in self.original_tree.find_data("code_element_struct"):
            for child in struct.find_data("decorator_list"):
                for decorator in child.find_data("identifier_def"):
                    if decorator.children[0] in ["contract_interface"]:
                        if tree in struct.iter_subtrees():
                            return

        # gather all decorators
        decorators = []
        for child in tree.children:
            if child.data == "decorator_list":
                for args in child.find_data("identifier_def"):
                    decorator = args.children[0]
                    decorators.append(str(decorator))

        for child in tree.children:
            if child.data == "identifier_def":
                function_name = str(child.children[0])
                break

        assert function_name

        func = FunctionType(function_name, getPosition(tree), self.fname, decorators)
        self.declared_functions.append(func)



# pylint: disable=unused-import
from amarna.rules.local_rules.ArithmeticOperationsRule import (
    MulArithmeticOperationsRule,
    DivArithmeticOperationsRule,
    AddArithmeticOperationsRule,
    SubArithmeticOperationsRule,
)
from amarna.rules.local_rules.UnusedArgumentRule import UnusedArgumentRule
from amarna.rules.local_rules.UnknownDecoratorRule import UnknownDecoratorRule
from amarna.rules.local_rules.UnusedImportRule import UnusedImportRule
from amarna.rules.local_rules.DeadStoreRule import DeadStoreRule
from amarna.rules.local_rules.UninitializedVariableRule import UninitializedVariableRule



from typing import Dict, List
from amarna.Result import Result, create_result_token

from amarna.rules.gatherer_rules.FunctionsReturningErrorsGatherer import (
    FunctionsReturningErrorsGatherer,
)
from amarna.rules.gatherer_rules.RValueFunctionCallsGatherer import RValueFunctionCallsGatherer

# pylint: disable=too-few-public-methods
class MustCheckReturnCodeRule:
    """
    Gather function calls and their return values.
    """

    RULE_TEXT = "This function returns an error code which must be properly checked."
    RULE_NAME = "must-check-error-code"

    def run_rule(self, gathered_data: Dict) -> List[Result]:
        functions_returning_errors = gathered_data[FunctionsReturningErrorsGatherer.GATHERER_NAME]
        function_calls = gathered_data[RValueFunctionCallsGatherer.GATHERER_NAME]

        results = []
        for call in function_calls:
            file_name, function_name, returned_list = call

            # when the function is one of those returning errors
            if function_name in functions_returning_errors:
                idx = functions_returning_errors[function_name]
                must_check = returned_list[idx].children[0]

                result = create_result_token(
                    file_name,
                    self.RULE_NAME,
                    self.RULE_TEXT,
                    must_check,
                )
                results.append(result)

            # also check if returned values were named success or exists
            else:
                for return_name in returned_list:
                    token = return_name.children[0]
                    if token in ["success", "exists"]:

                        result = create_result_token(
                            file_name,
                            "sucess-must-be-checked",
                            self.RULE_TEXT,
                            token,
                        )
                        results.append(result)

        return results



import re
from typing import Dict, List

from lark import Token
from amarna.Result import ResultMultiplePositions, result_multiple_positions, getPosition

from amarna.rules.gatherer_rules.AllFunctionCallsGatherer import (
    AllFunctionCallsGatherer,
    FunctionCallType,
)

UPPER_CASE_PATTERN = re.compile("^[A-Z_]{2,}$")


def is_constant_case(s: str) -> bool:
    return bool(UPPER_CASE_PATTERN.match(s))


# pylint: disable=too-few-public-methods,anomalous-backslash-in-string
class UniformAssertsConstants:
    """
    Look for different asserts where the same constant is used differently.
    assert_.*[A-Z_0-9]{6,}.*\)
    """

    RULE_TEXT = "This assertion uses the same constant differently [here](0) and [here](1)."
    RULE_NAME = "inconsistent-assert-constant"

    def run_rule(self, gathered_data: Dict) -> List[ResultMultiplePositions]:
        # pylint: disable=too-many-locals
        function_calls: List[FunctionCallType] = gathered_data[
            AllFunctionCallsGatherer.GATHERER_NAME
        ]

        results = []
        constant_uses = {}

        for call in function_calls:
            if "assert" in call.function_name:

                for arg_tree in call.arguments.children:
                    # get all tokens in the argument that are a constant
                    argument_tokens = sorted(
                        list(
                            arg_tree.scan_values(
                                lambda v: isinstance(v, Token) and is_constant_case(v.value)
                            )
                        )
                    )

                    if not argument_tokens:
                        continue

                    # create a dictionary key with the used constants
                    constants_key = "$".join(argument_tokens)

                    if constants_key not in constant_uses:
                        # when we haven't seen these constants being used,
                        # add their argument tree and the current filename to the dictionary
                        constant_uses[constants_key] = (arg_tree, call.arguments, call.file_name)

                    else:
                        old_tree, old_args, old_filename = constant_uses[constants_key]
                        # otherwise, when the trees differ it means that
                        # they are being used differently
                        if old_tree != arg_tree:

                            result = result_multiple_positions(
                                [call.file_name, old_filename],
                                self.RULE_NAME,
                                self.RULE_TEXT,
                                [getPosition(call.arguments), getPosition(old_args)],
                            )
                            results.append(result)

        return results



from typing import TypeVar, Generic

from lark import Tree, Visitor

# We need generic typing because classes derived from GenericGatherer
# Will return different types for gather
GenericGatherType = TypeVar("GenericGatherType")


class GenericGatherer(Visitor, Generic[GenericGatherType]):
    """
    Generic class for a rule that gathers data.
    """

    def gather(self, fname: str, tree: Tree) -> GenericGatherType:
        # TODO (montyly): investigate if this pylint rule should be removed
        # and an __init__ function created
        # pylint: disable=attribute-defined-outside-init
        self.fname = fname
        self.results: GenericGatherType = []  # type: ignore
        self.original_tree = tree
        self.visit(tree)
        return self.get_gathered_data()

    def get_gathered_data(self) -> GenericGatherType:
        # Overriden in inherited class
        # TODO (montyly): use abc?
        return self.results



from typing import Dict, List
from amarna.Result import ResultMultiplePositions, result_multiple_positions

from amarna.rules.gatherer_rules.DeclaredFunctionsGatherer import (
    DeclaredFunctionsGatherer,
    FunctionType,
)
from amarna.rules.gatherer_rules.ImportGatherer import ImportGatherer, ImportType


class ImplicitImportRule:
    """
    External, view and l1_handler declared functions will be imported even if not explicitly imported.
    """

    RULE_TEXT = "[This](0) function will be imported by [here](1), even though it was not explicitly imported."
    RULE_NAME = "implicit-import"

    DECORATORS = ["external", "view", "l1_handler"]

    def run_rule(self, gathered_data: Dict) -> List[ResultMultiplePositions]:
        declared_functions: List[FunctionType] = gathered_data[
            DeclaredFunctionsGatherer.GATHERER_NAME
        ]

        import_stmts: List[ImportType] = gathered_data[ImportGatherer.GATHERER_NAME]

        results = []

        for func in declared_functions:
            if any(decorator in self.DECORATORS for decorator in func.decorators):
                files_marked = []

                for imp in import_stmts:
                    # check if there is any import from the location
                    # of the external function
                    if func.file_location.endswith(imp.where_imported):
                        # do not flag the same file repeatedly
                        if imp.file_location in files_marked:
                            continue

                        files_marked.append(imp.file_location)

                        result = result_multiple_positions(
                            [func.file_location, imp.file_location],
                            self.RULE_NAME,
                            self.RULE_TEXT,
                            [func.position, imp.location],
                        )
                        results.append(result)

        return results



from typing import Dict, List
from amarna.Result import ResultMultiplePositions, result_multiple_positions
from amarna.rules.GenericRule import GenericRule

from amarna.rules.gatherer_rules.DeclaredFunctionsGatherer import (
    DeclaredFunctionsGatherer,
    FunctionType,
)
from amarna.rules.gatherer_rules.AllFunctionCallsGatherer import (
    AllFunctionCallsGatherer,
    FunctionCallType,
)


class UnenforcedViewRule(GenericRule):
    """
    Find state modifications in functions with @view decorator.
    """

    RULE_TEXT = (
        "[This](0) function call modifies state but is called within [this](1) @view context."
    )
    RULE_NAME = "unenforced-view"

    def run_rule(self, gathered_data: Dict) -> List[ResultMultiplePositions]:
        declared_functions: List[FunctionType] = gathered_data[
            DeclaredFunctionsGatherer.GATHERER_NAME
        ]
        function_calls: List[FunctionCallType] = gathered_data[
            AllFunctionCallsGatherer.GATHERER_NAME
        ]

        results = []

        def check_parents(call: FunctionCallType, original_call: FunctionCallType):
            for func in declared_functions:

                # if it is not the parent, skip it
                if func.name != call.parent_function:
                    continue

                # if the parent is a view, add to result
                if "view" in func.decorators:
                    result = result_multiple_positions(
                        [original_call.file_name, func.file_location],
                        self.RULE_NAME,
                        self.RULE_TEXT,
                        [original_call.position, func.position],
                    )
                    if result.position_list not in (r.position_list for r in results):
                        results.append(result)
                    break

                seen = [call.function_name]
                # check if the parent is called somewhere else
                for f in function_calls:
                    if f.function_name != call.parent_function:
                        continue
                    if f.function_name not in seen:
                        seen.append(f.function_name)
                        check_parents(f, original_call)

        for f in function_calls:
            if (
                f.tail_name == "write"
                or f.tail_name == "emit"
                or f.function_name == "storage_write"
            ):
                check_parents(f, f)

        return results



from lark import Tree

from amarna.rules.GenericRule import GenericRule
from amarna.Result import PositionType, create_result


class UnknownDecoratorRule(GenericRule):
    """
    Check for misspelled function decorators.
    """

    RULE_TEXT = "Unknown of misspelled function decorator."
    RULE_NAME = "unknown-decorator"

    known_decorators = [
        "storage_var",
        "external",
        "view",
        "constructor",
        "l1_handler",
        "known_ap_change",
        "event",
        "raw_input",
        "raw_output",
    ]

    def code_element_function(self, tree: Tree) -> None:
        unknown_decorators = []
        for child in tree.children:
            if child.data == "decorator_list":
                for args in child.find_data("identifier_def"):
                    decorator = args.children[0]
                    if decorator not in self.known_decorators:
                        unknown_decorators.append(decorator)

        for arg in unknown_decorators:
            # TODO (montyly): mypy compain about the next attributes access
            positions = PositionType(arg.line, arg.column, arg.end_line, arg.end_column)  # type: ignore
            sarif = create_result(
                self.fname,
                self.RULE_NAME,
                self.RULE_TEXT,
                positions,
            )
            self.results.append(sarif)



from typing import Set
from lark import Tree, Token
from amarna.Result import PositionType, create_result

from amarna.rules.GenericRule import GenericRule


class UnusedImportRule(GenericRule):
    """
    Check for unused imports.
    """

    RULE_TEXT = "Unused imports could be removed."
    RULE_NAME = "unused-imports"

    # visit the whole cairo_file node of the AST
    def cairo_file(self, tree: Tree) -> None:
        # pylint: disable=too-many-locals,too-many-branches

        imports = set()
        # iterate over the imports
        for code_imports in tree.find_data("code_element_import"):
            for child in code_imports.children:

                if child.data == "aliased_identifier":

                    # this is just a simple import
                    if len(child.children) == 1:
                        child_id = child.children[0].children[0]
                        if child_id in imports:
                            print("double import of ", child_id)
                        imports.add(child_id)

                    # this is an aliased import import XX as YY, we keep the new name
                    elif len(child.children) == 2:
                        child_id = child.children[1].children[0]
                        if child_id in imports:
                            print("double import of ", child_id)
                        imports.add(child.children[1].children[0])

        used_ids: Set[Token] = set()
        # gather identifiers used in the code
        for function_code in self.original_tree.find_data("code_element_function"):
            for code_child in function_code.find_data("identifier"):
                # TODO (montyly): attribute access error
                used_ids.add(code_child.children[0])  # type: ignore

        # gather types used in struct declaractions
        for struct_declration in self.original_tree.find_data("code_element_struct"):
            for code_child in struct_declration.find_data("identifier"):
                # TODO (montyly): attribute access error
                used_ids.add(code_child.children[0])  # type: ignore

        unused_imports = imports - used_ids
        if not unused_imports:
            return

        # gather all hint code and check if the imports are there
        all_hints = ""
        for hint in self.original_tree.find_data("code_element_hint"):
            all_hints += hint.children[0]

        # remove imports used in hints
        used_in_hints = set()
        for unused in unused_imports:
            if unused.value in all_hints:
                used_in_hints.add(unused)

        unused_imports = unused_imports - used_in_hints

        # if unused_imports:
        #     print(f"In file {self.fname}:")

        # report unused imports
        for arg in sorted(unused_imports):
            # print(f"\t{arg.value} imported at line {arg.line}")
            positions = PositionType(arg.line, arg.column, arg.end_line, arg.end_column)  # type: ignore
            sarif = create_result(
                self.fname,
                self.RULE_NAME,
                self.RULE_TEXT,
                positions,
            )
            self.results.append(sarif)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="amarna",
    version="0.1.2",
    description="Amarna is a static-analyzer for the Cairo programming language.",
    url="https://github.com/crytic/amarna",
    author="Trail of Bits",
    license="AGPL-3.0",
    packages=find_packages(),
    package_data={"amarna": ["grammars/cairo.lark"]},
    install_requires=[
        "lark>=1.1.2",
        "pydot>=1.4.2",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "amarna=amarna.command_line:main",
        ],
    },
)



from dataclasses import dataclass
from typing import List

from lark import Tree
from amarna.Result import PositionType, getPosition

from amarna.rules.GenericGatherer import GenericGatherer


@dataclass
class CallbackFunctionType:
    """Represents a function callback."""

    file_name: str
    function_name: str
    position: PositionType


class FunctionsUsedAsCallbacksGatherer(GenericGatherer):
    """
    Gather all functions that are used as callbacks.
    """

    GATHERER_NAME = "FunctionsUsedAsCallbacks"

    def __init__(self) -> None:
        super().__init__()
        self.function_calls: List[CallbackFunctionType] = []

    def get_gathered_data(self) -> List[CallbackFunctionType]:
        return self.function_calls

    def function_call(self, tree: Tree) -> None:

        children_id = tree.children[0]
        # TODO (montyly): attribute access error
        function_name = children_id.children[0].value  # type: ignore
        if function_name == "serialize_array":
            arguments = tree.children[1]
            callback_arg = arguments.children[-1]
            ids = list(callback_arg.find_data("identifier"))
            # variable callback
            if len(ids) == 1:
                token = ids[0].children[0]

            # manually callback such as asset_config_hash_serialize + __pc__ - ret_pc_label
            # TODO: check ast for name + __pc__ - ret_pc_label
            elif len(ids) == 3:
                # also idx 0
                token = ids[0].children[0]

            else:
                raise Exception(
                    "Function call path not implemented in amarna, please open an issue"
                )

            # add to the gatherer data
            callback = CallbackFunctionType(self.fname, str(token), getPosition(tree))
            self.function_calls.append(callback)

        elif function_name == "get_label_location":
            for identifier in tree.find_data("atom_identifier"):
                token = identifier.children[0].children[0]

                # add to the gatherer data
                callback = CallbackFunctionType(self.fname, str(token), getPosition(tree))
                self.function_calls.append(callback)



import re
from typing import Set

from lark import Tree, Token
from amarna.rules.GenericRule import GenericRule
from amarna.Result import create_result_token


class UninitializedVariableRule(GenericRule):
    """
    Check for uninitialized local variables.
    """

    RULE_TEXT = "This local variable is uninitialized and never assigned."
    RULE_NAME = "uninitialized-variable"

    def code_element_function(self, tree: Tree) -> None:
        local_variables: Set[Token] = set()
        assigned_variables: Set[Token] = set()

        for local_declaration in tree.find_data("code_element_local_var"):
            # only gather unintialized locals
            if len(local_declaration.children) == 1:
                for child in local_declaration.children[0].find_data("identifier_def"):
                    local_variables.add(child.children[0])

        # ignore if they are assigned
        for let_reference in tree.find_data("code_element_reference"):
            for child in let_reference.children[0].find_data("identifier_def"):
                assigned_variables.add(child.children[0])

        # ignore if they are expressed in a inst_assert_eq used as assignment
        for a in tree.find_data("inst_assert_eq"):
            for children_id in a.children[0].find_data("identifier"):
                assigned_variables.add(children_id.children[0])

        # ignore if they are expressed in a assert used as assignment
        for a in tree.find_data("code_element_compound_assert_eq"):
            for b in a.children:
                if b.data == "atom_identifier":
                    assigned_variables.add(b.children[0].children[0])

        uninitialized_locals = local_variables - assigned_variables
        if not uninitialized_locals:
            return

        # gather all hint code and check if the variables are there
        all_hints = ""
        for hint in self.original_tree.find_data("code_element_hint"):
            all_hints += hint.children[0]

        # for now just ignore if they're in a hint
        used_in_hints = set()
        for uninitialized in uninitialized_locals:
            if re.search("ids\." + uninitialized.value + "( |,|\.).*=", all_hints):
                used_in_hints.add(uninitialized)

        uninitialized_locals = uninitialized_locals - used_in_hints

        for tok in uninitialized_locals:
            result = create_result_token(
                self.fname,
                self.RULE_NAME,
                self.RULE_TEXT,
                tok,
            )
            self.results.append(result)



from typing import Dict, List
from amarna.Result import Result, create_result_token

from amarna.rules.gatherer_rules.RValueFunctionCallsGatherer import (
    RValueFunctionCallsGatherer,
)

# pylint: disable=too-few-public-methods
class MustCheckOverflow:
    """
    Gather function calls and their return values.
    """

    RULE_TEXT = (
        "This function returns a variable that indicates"
        + " if an overflow occurred and must be properly checked."
    )
    IGNORED_RULE_TEXT = "This function ignores the overflow indicator variable."

    RULE_NAME = "must-check-overflow"

    # dictionary with function_name : index of the returned overflow variable
    OVERFLOW_FUNCTIONS = {"uint256_add": 1}

    def run_rule(self, gathered_data: Dict) -> List[Result]:

        function_calls = gathered_data[RValueFunctionCallsGatherer.GATHERER_NAME]

        results = []
        for call in function_calls:
            file_name, function_name, returned_list = call

            # when the function is one of those returning errors
            if function_name in self.OVERFLOW_FUNCTIONS:
                overflow_idx = self.OVERFLOW_FUNCTIONS[function_name]

                must_check = returned_list[overflow_idx].children[0]

                if must_check == "_":
                    # the overflow return value is being ignored
                    rule = self.IGNORED_RULE_TEXT
                else:
                    # here we say that it must be checked
                    rule = self.RULE_TEXT

                result = create_result_token(
                    file_name,
                    self.RULE_NAME,
                    rule,
                    must_check,
                )
                results.append(result)

        return results



from typing import List, Union
from dataclasses import dataclass

from lark import tree, Tree
from amarna.Result import PositionType, getPosition
from amarna.rules.GenericGatherer import GenericGatherer


@dataclass
class FunctionCallType:
    """Represents a function call."""

    file_name: str
    parent_function: str
    function_name: str
    tail_name: str
    position: PositionType
    arguments: Union[str, tree.Tree]


class AllFunctionCallsGatherer(GenericGatherer[list]):
    """
    Gather all function calls independently of being an rvalue or a tail-call.
    """

    GATHERER_NAME = "AllFunctionCallsGatherer"

    def __init__(self) -> None:
        super().__init__()
        self.function_calls: List[FunctionCallType] = []

    def get_gathered_data(self) -> List[FunctionCallType]:
        return self.function_calls

    def code_element_function(self, tree: Tree) -> None:
        parent_name: Optional[str] = None

        for child in tree.children:
            if child.data == "identifier_def":
                parent_name = str(child.children[0])
                break

        assert parent_name

        for func in tree.find_data("function_call"):
            func_id = func.children[0]
            # TODO (montyly): Mypy complains about .value not be present
            function_name: str = func_id.children[0].value  # type: ignore
            if len(func_id.children) > 1:
                tail_name: str = func_id.children[1].value  # type: ignore
            else:
                tail_name: str = None
            arguments = func.children[-1]

            function_call = FunctionCallType(
                self.fname, parent_name, function_name, tail_name, getPosition(func), arguments
            )
            self.function_calls.append(function_call)



# pylint: disable=unused-import,line-too-long
from amarna.rules.gatherer_rules.FunctionsReturningErrorsGatherer import (
    FunctionsReturningErrorsGatherer,
)
from amarna.rules.gatherer_rules.DeclaredFunctionsGatherer import DeclaredFunctionsGatherer
from amarna.rules.gatherer_rules.AllFunctionCallsGatherer import AllFunctionCallsGatherer
from amarna.rules.gatherer_rules.RValueFunctionCallsGatherer import RValueFunctionCallsGatherer
from amarna.rules.gatherer_rules.FunctionsUsedAsCallbacksGatherer import (
    FunctionsUsedAsCallbacksGatherer,
)
from amarna.rules.gatherer_rules.ImportGatherer import ImportGatherer
from amarna.rules.gatherer_rules.FunctionsEmittingEventsGatherer import (
    FunctionsEmittingEventsGatherer,
)



from typing import Set

from lark import Tree, Token
from amarna.rules.GenericRule import GenericRule
from amarna.Result import create_result_token


class DeadStoreRule(GenericRule):
    """
    Check for dead dead_stores.
    """

    RULE_TEXT = "This variable is assigned or declared here but not used before a return statement."
    RULE_NAME = "dead-store"

    def code_element_function(self, tree: Tree) -> None:
        # pylint: disable=too-many-branches,too-many-nested-blocks,too-many-locals
        # gather implicit arguments
        implicits_and_arguments = []
        for impls in tree.find_data("implicit_arguments"):
            for args in impls.find_data("identifier_def"):
                implicits_and_arguments.append(str(args.children[0]))

        # gather argument names
        for allargs in tree.find_data("arguments"):
            for args in allargs.find_data("identifier_def"):
                implicits_and_arguments.append(str(args.children[0]))

        # gather call template names
        for allargs in tree.find_data("non_def_expr_assignment"):
            for args in allargs.find_data("identifier_def"):
                implicits_and_arguments.append(str(args.children[0]))

        # gather with_attr names
        for withattr in tree.find_data("code_element_with_attr"):
            for args in withattr.find_data("identifier_def"):
                implicits_and_arguments.append(str(args.children[0]))

        defines: Set[Token] = set()
        for main_child in tree.children:
            if main_child.data != "code_block":
                continue

            for child in main_child.children:
                if child.children[0].data not in ["code_element_struct"]:
                    # add identifier definitions that aren't _, __fp__ or pc_val
                    for defn in child.find_data("identifier_def"):
                        if defn.children[0] not in ["_", "__fp__", "pc_val"]:
                            defines.add(defn.children[0])  # type: ignore

                    # remove identifier uses
                    for uses in child.find_data("identifier"):
                        tok = uses.children[0]
                        if tok in defines:
                            defines.remove(tok)  # type: ignore

                    # add lvalues. These can be false positives
                    # when a = b is an assert and not an assignment
                    for a in child.find_data("inst_assert_eq"):
                        for children_id in a.children[0].find_data("identifier"):
                            defines.add(children_id.children[0])  # type: ignore

                    # in a return statement, check which variables were not used
                    for subcode in child.children[0].find_data("code_element_return"):
                        tokens = [
                            str(tok) for tok in subcode.scan_values(lambda v: isinstance(v, Token))
                        ]
                        for dead_store in sorted(defines):
                            if dead_store in tokens or dead_store in implicits_and_arguments:
                                continue
                            sarif = create_result_token(
                                self.fname,
                                self.RULE_NAME,
                                self.RULE_TEXT,
                                dead_store,
                            )
                            if sarif in self.results:
                                continue
                            self.results.append(sarif)



from dataclasses import dataclass
from typing import Any, Optional, Dict, Union
from typing import List
import os
from lark import Tree, Token
import json


@dataclass
class PositionType:
    """Tracks the position in a file"""

    start_line: int
    start_col: int
    end_line: int
    end_col: int


class Result:
    def __init__(self, filename: str, rule_name: str, text: str, position: PositionType) -> None:
        self.filename = filename
        self.rule_name = rule_name
        self.text = text
        self.position = position

    def to_sarif(self) -> Dict:
        """
        Return a SARIF dictionary for a filename, rule, text description and code location.
        """
        return {
            "ruleId": self.rule_name,
            "level": "warning",
            "message": {"text": self.text},
            "locations": [
                {
                    "physicalLocation": {
                        "artifactLocation": {"uri": "file://" + self.filename, "index": 0},
                        "region": sarif_region_from_position(self.position),
                    }
                }
            ],
        }

    def to_summary(self) -> str:
        short_name = os.path.basename(self.filename)
        return f"[{self.rule_name}] in {short_name}:{self.position.start_line}:{self.position.start_col}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Result):
            return (
                self.filename == other.filename
                and self.position == other.position
                and self.rule_name == other.rule_name
            )
        return False


class ResultMultiplePositions:
    def __init__(
        self,
        filenames: str,
        rule_name: str,
        text: str,
        position_list: List[PositionType],
    ) -> None:
        assert len(filenames) == len(position_list)
        self.filenames = filenames
        self.rule_name = rule_name
        self.text = text
        self.position_list = position_list

    def to_sarif(self) -> Dict:
        related_location_list = [
            {
                "id": idx,
                "physicalLocation": {
                    "artifactLocation": {"uri": "file://" + filename, "index": 0},
                    "region": sarif_region_from_position(self.position_list[idx]),
                },
            }
            for idx, filename in enumerate(self.filenames)
        ]
        return {
            "ruleId": self.rule_name,
            "level": "warning",
            "message": {"text": self.text},
            "locations": [
                {
                    "physicalLocation": {
                        "artifactLocation": {"uri": "file://" + self.filenames[0], "index": 0},
                        "region": sarif_region_from_position(self.position_list[0]),
                    },
                }
            ],
            "relatedLocations": related_location_list,
        }

    def to_summary(self) -> str:
        interpolated = [
            f"{os.path.basename(loc)}:{res.start_line}:{res.start_col}"
            for loc, res in zip(self.filenames, self.position_list)
        ]
        if len(interpolated) > 1:
            prefix = " and "
        else:
            prefix = ""

        full_str = ", ".join(interpolated[:-1]) + f"{prefix}{interpolated[-1]}"

        return f"[{self.rule_name}] {self.text} in " + full_str


ResultTypes = Union[Result, ResultMultiplePositions]

SARIF_MODE = "sarif_mode"
SUMMARY_MODE = "summary_mode"


def output_result(results: List[ResultTypes], fname: str, print_output: bool, mode: str) -> None:
    if mode == SARIF_MODE:
        create_sarif(results, fname, print_output)

    elif mode == SUMMARY_MODE:
        summary = create_summary(results)
        if print_output:
            print(summary)

        if fname:
            with open(os.path.join(os.getcwd(), fname), "w", encoding="utf8") as f:
                f.write(summary)


def create_summary(results: List[ResultTypes]) -> str:
    return "\n".join(map(lambda res: res.to_summary(), results))


def create_result(filename: str, rule_name: str, text: str, position: PositionType) -> Result:
    return Result(filename, rule_name, text, position)


def create_result_token(filename: str, rule_name: str, text: str, token: Token) -> Result:
    return Result(filename, rule_name, text, token_positions(token))


def result_multiple_positions(
    filenames: str,
    rule_name: str,
    text: str,
    position_list: List[PositionType],
) -> ResultMultiplePositions:
    return ResultMultiplePositions(filenames, rule_name, text, position_list)


def create_sarif(
    results: List[Any], fname: Optional[str] = None, printoutput: bool = False
) -> None:
    """
    Create the sarif output json for the results, and write it to file or print it.
    """
    results_sarif = list(map(lambda res: res.to_sarif(), results))
    sarif = {
        "version": "2.1.0",
        "$schema": "http://json.schemastore.org/sarif-2.1.0-rtm.4",
        "runs": [{"tool": {"driver": {"name": "Amarna"}}, "results": results_sarif}],
    }

    if fname:
        with open(os.path.join(os.getcwd(), fname), "w", encoding="utf8") as f:
            json.dump(sarif, f, indent=1)

    if printoutput:
        print(json.dumps(sarif))


def sarif_region_from_position(position: PositionType) -> Dict[str, int]:
    """
    Return the sarif region field for a code location
    """
    return {
        "startLine": position.start_line,
        "startColumn": position.start_col,
        "endLine": position.end_line,
        "endColumn": position.end_col,
    }


SarifOuputType = Dict[str, Any]


def token_positions(token: Token) -> PositionType:
    """
    Get the file locations of a token: line, col, end_line, end_col
    """
    return PositionType(
        token.line,
        token.column,
        token.end_line,
        token.end_column,
    )


def getPosition(tree: Tree) -> PositionType:
    """
    Get the file locations of the tree: line, col, end_line, end_col
    """
    meta = tree.meta
    return PositionType(meta.line, meta.column, meta.end_line, meta.end_column)
